@@ -14,6 +14,7 @@ import haiku as hk
 import jax
 import jax.numpy as jnp
 import jmp
+import numpy as np
 import optax
 import tensorflow as tf
 from chex import PRNGKey, ArrayTree, Scalar, Array
@@ -85,7 +86,7 @@ class Trainer:
         validation_fn: Optional[LossFunction] = None,
         trainable_predicate: Callable[[str, str, Array], bool] = None,
         run_eagerly: bool = False,
-        num_devices: Optional[int] = None,
+        num_devices: int = 1,
         mp_policy: Optional[jmp.Policy] = None,
         skip_nonfinite_updates: bool = False,
         loss_scale: Optional[jmp.LossScale] = None,
@@ -116,7 +117,6 @@ class Trainer:
             )
 
         local_device_count = jax.local_device_count()
-        num_devices = num_devices or local_device_count
         if num_devices > local_device_count:
             raise ValueError(
                 f"num_devices is {num_devices}, but there are only "
@@ -172,11 +172,13 @@ class Trainer:
             self._partitioned_loss, has_aux=True
         )(trainable_params, non_trainable_params, train_state.state, key, step, batch)
         grads = self._mp_policy.cast_to_compute(grads)
-        grads = train_state.loss_scale.unscale(grads)
 
         if self._num_devices > 1:
             grads = jax.lax.pmean(grads, axis_name=self.cross_replica_axis)
 
+        grads = train_state.loss_scale.unscale(grads)
+        grad_norm = optax.global_norm(grads)
+        aux["gradient_norm"] = grad_norm
         grads = self._mp_policy.cast_to_param(grads)
 
         updates, new_opt_state = self._optimizer.update(
@@ -195,12 +197,12 @@ class Trainer:
             )
             aux["mp_grads_finite"] = grads_finite
 
-        if self._mp_policy is not None:
+        if not isinstance(self._loss_scale, jmp.NoOpLossScale):
             aux["mp_loss_scale"] = loss_scale.loss_scale
 
-        should_skip = False
+        should_skip = jnp.zeros(grad_norm.shape, jnp.bool_)
         if self._gradient_skipping_threshold is not None:
-            should_skip = optax.global_norm(grads) > self._gradient_skipping_threshold
+            should_skip = grad_norm > self._gradient_skipping_threshold
             new_trainable_params, new_state, new_opt_state = jmp.select_tree(
                 should_skip,
                 (trainable_params, train_state.state, train_state.opt_state),
@@ -217,12 +219,15 @@ class Trainer:
             aux = jax.lax.pmean(aux, axis_name=self.cross_replica_axis)
             loss = jax.lax.pmean(loss, axis_name=self.cross_replica_axis)
 
-        if self._ema_rate is not None and not should_skip:
+        if self._ema_rate is not None:
             new_ema_params = jax.tree_multimap(
                 lambda e, p: e * self._ema_rate + (1 - self._ema_rate) * p,
                 train_state.ema_params,
                 new_params,
             )
+            new_ema_params = jmp.select_tree(
+                should_skip, train_state.ema_params, new_ema_params
+            )
         else:
             new_ema_params = train_state.ema_params
 
@@ -332,7 +337,7 @@ class Trainer:
             byte_size = hk.data_structures.tree_bytes(params)
             num_params = hk.data_structures.tree_size(params)
 
-        print(f"Number of Parameters: {num_params}, {byte_size / 1e6:.2f}MB")
+        print(f"Total Parameters: {num_params}, {byte_size / 1e6:.2f}MB")
 
         metrics = defaultdict(Mean)
 
@@ -362,6 +367,12 @@ class Trainer:
             aux = jax.device_get(aux)
             loss = jax.device_get(loss)
 
+            train_step_logs = {k: np.asarray(v).item() for k, v in aux.items()}
+            train_step_logs["loss"] = np.asarray(loss).item()
+
+            for callback in callbacks:
+                callback.on_train_step(step, train_step_logs)
+
             metrics["loss"].update_state(loss)
             for k, v in aux.items():
                 metrics[k].update_state(v)
@@ -379,6 +390,11 @@ class Trainer:
                         val_batch,
                     )
 
+                    for callback in callbacks:
+                        callback.on_validation_step(
+                            train_state, self._get_pmap_keys(), val_batch
+                        )
+
                     if self._num_devices > 1:
                         val_aux = jax.tree_map(lambda x: x[0], val_aux)
                         val_loss = jax.tree_map(lambda x: x[0], val_loss)
